/// <summary>
/// Some of the code and ideas in this class were taken from a paper entitled
/// "Fast and Simple Physics using Sequential Impulses" by Erin Catto
/// </summary>

using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using Microsoft.Xna.Framework;

namespace KLib.Physics.Collisions
{

    public class Arbiter2 : IEquatable<Arbiter2> 
    {
        private PhysicsManager physicsSimulator;
        private float frictionCoefficientCombined;
        
        internal Geom geomA;
        internal Geom geomB;

        internal RigidBody bodyA;
        internal RigidBody bodyB;

        internal ContactList contactList;
        private ContactList newContactList;
        private ContactList mergedContactList;

        //used for ref/out methods
        private Vector2 vec1 = Vector2.Zero;
        private Vector2 vec2 = Vector2.Zero;
        private Vector2 vec3 = Vector2.Zero;
        private Vector2 vec4 = Vector2.Zero;
        private Vector2 vec5 = Vector2.Zero;
        private float float1 = 0;
        private float float2 = 0;

        private Comparison<Contact> contactComparer;

        public Arbiter2() { }

        internal Arbiter2(Geom geom1, Geom geom2, PhysicsManager physicsSimulator) {
            ConstructArbiter(geom1, geom2, physicsSimulator);
        }

        internal void ConstructArbiter(Geom geom1, Geom geom2, PhysicsManager physicsSimulator)
        {
            this.physicsSimulator = physicsSimulator;

            this.GeomA = geom1;
            this.GeomB = geom2;

            this.bodyA = geom1.Body;
            this.bodyB = geom2.Body;

            this.frictionCoefficientCombined = (bodyA.FrictionCoefficient + bodyB.FrictionCoefficient) / 2f;
            
            InitializeContactLists(physicsSimulator.maxContactsToDetect);

            //to prevent garbage build up the contact comparer is pre-initialized and held in memory
            contactComparer = new Comparison<Contact>(CompareSeperation);
        }

        public Geom GeomA {
            get { return geomA; }
            set { geomA = value; }
        }

        public Geom GeomB {
            get { return geomB; }
            set { geomB = value; }
        }

        #region PreStepImpulse variables
        Vector2 r1;
        Vector2 r2;
        float rn1;
        float rn2;
        float rt1;
        float rt2;
        float kTangent;
        float min;
        float restitution;
        float kNormal;
        #endregion
        internal void PreStepImpulse(float inverseDt) {
            Contact contact;
            if (!geomA.CollisionResponseEnabled || !geomB.CollisionResponseEnabled) { return; }
            for (int i = 0; i < contactList.Count; i++) {
                contact = contactList[i];

                //calculate contact offset from body position
                Vector2.Subtract(ref contact.Position, ref bodyA.position, out r1);
                Vector2.Subtract(ref contact.Position, ref bodyB.position, out r2);

                //project normal onto offset vectors
                Vector2.Dot(ref r1, ref contact.Normal, out rn1);
                Vector2.Dot(ref r2, ref contact.Normal, out rn2);

                //calculate mass normal
                float invMassSum = bodyA.InverseMass + bodyB.InverseMass;
                Vector2.Dot(ref r1, ref r1, out float1);
                Vector2.Dot(ref r2, ref r2, out float2);
                kNormal = invMassSum + bodyA.InverseMomentOfIntertia * (float1 - rn1 * rn1) + bodyB.InverseMomentOfIntertia * (float2 - rn2 * rn2);
                contact.MassNormal = 1f / kNormal;

                //calculate mass tangent
                float1 = 1;
                Utilities.Math.Cross(ref contact.Normal, ref float1, out tangent);
                Vector2.Dot(ref r1, ref tangent, out rt1);
                Vector2.Dot(ref r2, ref tangent, out rt2);
                kTangent = bodyA.InverseMass + bodyB.InverseMass;

                Vector2.Dot(ref r1, ref r1, out float1);
                Vector2.Dot(ref r2, ref r2, out float2);
                kTangent += bodyA.InverseMomentOfIntertia * (float1 - rt1 * rt1) + bodyB.InverseMomentOfIntertia * (float2 - rt2 * rt2);
                contact.MassTangent = 1f / kTangent;

                //calc velocity bias
                min = Math.Min(0, physicsSimulator.AllowedPenitration + contact.Separation);
                contact.NormalVelocityBias = -physicsSimulator.BiasFactor * inverseDt * min;

                //Compute the restitution, we average the restitution of the two bodies
                //restitution = (2.0f + _geometryA.RestitutionCoefficient + _geometryB.RestitutionCoefficient) * 0.5f;
                restitution = geomA.RestitutionCoefficient + geomB.RestitutionCoefficient * .5f;

                //calc bounce velocity
                bodyA.GetVelocityAtWorldOffset(ref r1, out vec1);
                bodyB.GetVelocityAtWorldOffset(ref r2, out vec2);
                Vector2.Subtract(ref vec2, ref vec1, out dv);

                //calc velocity difference along contact normal
                Vector2.Dot(ref dv, ref contact.Normal, out vn);
                contact.BounceVelocity = vn * restitution;

                //apply accumulated impulse
                Vector2.Multiply(ref contact.Normal, contact.NormalImpulse, out vec1);
                Vector2.Multiply(ref tangent, contact.TangentImpulse, out vec2);
                Vector2.Add(ref vec1, ref vec2, out impulse);

                bodyB.ApplyImpulseAtWorldOffset(ref impulse, ref r2);

                Vector2.Multiply(ref impulse, -1, out impulse);
                bodyA.ApplyImpulseAtWorldOffset(ref impulse, ref r1);

                contact.NormalImpulseBias = 0;
                contactList[i] = contact;
            }
        }

        #region variables for ApplyImpulse
        Contact contact;
        Vector2 tangent = Vector2.Zero;
        Vector2 dv = Vector2.Zero;

        float vn = 0;
        float normalVelocityBias = 0;
        float normalImpulseBias = 0;
        float normalImpulseBiasOriginal = 0;

        float normalImpulse = 0;
        float oldNormalImpulse = 0;
        float maxTangentImpulse = 0;

        float vt = 0;
        float tangentImpulse = 0;
        float oldTangentImpulse = 0;

        Vector2 impulse = Vector2.Zero;
        Vector2 impulseBias = Vector2.Zero;
        #endregion
        internal void ApplyImpulse() {
            if (!geomA.CollisionResponseEnabled || !geomB.CollisionResponseEnabled) { return; }
            for (int i = 0; i < contactList.Count; i++) {

                contact = contactList[i];
                                
                #region INLINE: Vector2.Subtract(ref contact.Position, ref bodyA.position, out contact.R1);
                contact.R1.X = contact.Position.X - bodyA.Position.X;
                contact.R1.Y = contact.Position.Y - bodyA.Position.Y;
                #endregion
                
                #region INLINE: Vector2.Subtract(ref contact.Position, ref bodyB.position, out contact.R2);
                contact.R2.X = contact.Position.X - bodyB.Position.X;
                contact.R2.Y = contact.Position.Y - bodyB.Position.Y;
                #endregion

                //calc velocity difference
                bodyA.GetVelocityAtWorldOffset(ref contact.R1, out vec1);
                bodyB.GetVelocityAtWorldOffset(ref contact.R2, out vec2);
                
                #region INLINE: Vector2.Subtract(ref vec2, ref vec1, out dv);
                                dv.X = vec2.X - vec1.X;
                                dv.Y = vec2.Y - vec1.Y;
                #endregion

                //calc velocity difference along contact normal
                #region INLINE: Vector2.Dot(ref dv, ref contact.Normal, out vn);
                vn = (dv.X * contact.Normal.X) + (dv.Y * contact.Normal.Y);
                #endregion

                                ////float normalImpulse = contact.MassNormal * (-vn + contact.NormalVelocityBias); //comment for preserve momentum
                normalImpulse = contact.MassNormal * -(vn + contact.BounceVelocity); //uncomment for preserve momentum

                //clamp accumulated impulse
                oldNormalImpulse = contact.NormalImpulse;
                contact.NormalImpulse = Math.Max(oldNormalImpulse + normalImpulse, 0);
                normalImpulse = contact.NormalImpulse - oldNormalImpulse;

                //apply contact impulse
                #region INLINE: Vector2.Multiply(ref contact.Normal, normalImpulse, out impulse);
                impulse.X = contact.Normal.X * normalImpulse;
                impulse.Y = contact.Normal.Y * normalImpulse;
                #endregion

                bodyB.ApplyImpulseAtWorldOffset(ref impulse, ref contact.R2);
                
                #region INLINE: Vector2.Multiply(ref impulse, -1, out impulse);
                impulse.X = impulse.X * -1;
                impulse.Y = impulse.Y * -1;
                #endregion

                bodyA.ApplyImpulseAtWorldOffset(ref impulse, ref contact.R1);

                //calc velocity bias difference (bias preserves momentum)
                bodyA.GetVelocityBiasAtWorldOffset(ref contact.R1, out vec1);
                bodyB.GetVelocityBiasAtWorldOffset(ref contact.R2, out vec2);
               
                #region INLINE: Vector2.Subtract(ref vec2, ref vec1, out dv);
                dv.X = vec2.X - vec1.X;
                dv.Y = vec2.Y - vec1.Y;
                #endregion

                //calc velocity bias along contact normal
                #region INLINE: Vector2.Dot(ref dv, ref contact.Normal, out normalVelocityBias);
                normalVelocityBias = (dv.X * contact.Normal.X) + (dv.Y * contact.Normal.Y);
                #endregion

                normalImpulseBias = contact.MassNormal * (-normalVelocityBias + contact.NormalVelocityBias);
                normalImpulseBiasOriginal = contact.NormalImpulseBias;
                contact.NormalImpulseBias = Math.Max(normalImpulseBiasOriginal + normalImpulseBias, 0);
                normalImpulseBias = contact.NormalImpulseBias - normalImpulseBiasOriginal;

                #region INLINE: Vector2.Multiply(ref contact.Normal, normalImpulseBias, out impulseBias);
                impulseBias.X = contact.Normal.X * normalImpulseBias;
                impulseBias.Y = contact.Normal.Y * normalImpulseBias;
                #endregion

                //apply bias impulse
                bodyB.ApplyImpulseBiasAtWorldOffset(ref impulseBias, ref contact.R2);

                #region INLINE: Vector2.Multiply(ref impulseBias, -1, out impulseBias);
                impulseBias.X = impulseBias.X * -1;
                impulseBias.Y = impulseBias.Y * -1;
                #endregion

                bodyA.ApplyImpulseBiasAtWorldOffset(ref impulseBias, ref contact.R1);

                //calc relative velocity at contact.
                bodyA.GetVelocityAtWorldOffset(ref contact.R1, out vec1);
                bodyB.GetVelocityAtWorldOffset(ref contact.R2, out vec2);

                #region INLINE: Vector2.Subtract(ref vec2, ref vec1, out dv);
                dv.X = vec2.X - vec1.X;
                dv.Y = vec2.Y - vec1.Y;
                #endregion

                //compute friction impulse
                maxTangentImpulse = frictionCoefficientCombined * contact.NormalImpulse;

                //Vector2 tangent = Calculator.Cross(contact.Normal, 1);
                float1 = 1;

                #region INLINE: Calculator.Cross(ref contact.Normal, ref float1, out tangent);
                tangent.X = float1 * contact.Normal.Y;
                tangent.Y = -float1 * contact.Normal.X;
                #endregion

                //float vt = Vector2.Dot(dv, tangent);
                #region INLINE: Vector2.Dot(ref dv, ref tangent, out vt);
                vt = (dv.X * tangent.X) + (dv.Y * tangent.Y);
                #endregion

                tangentImpulse = contact.MassTangent * (-vt);

                ////clamp friction
                oldTangentImpulse = contact.TangentImpulse;
                contact.TangentImpulse = MathHelper.Clamp(oldTangentImpulse + tangentImpulse, -maxTangentImpulse, maxTangentImpulse);
                tangentImpulse = contact.TangentImpulse - oldTangentImpulse;

                //apply friction impulse
                #region INLINE:Vector2.Multiply(ref tangent, tangentImpulse, out impulse);
                impulse.X = tangent.X * tangentImpulse;
                impulse.Y = tangent.Y * tangentImpulse;
                #endregion

                //apply impulse
                bodyB.ApplyImpulseAtWorldOffset(ref impulse, ref contact.R2);
                #region INLINE: Vector2.Multiply(ref impulse, -1, out impulse);
                impulse.X = impulse.X * -1;
                impulse.Y = impulse.Y * -1;
                #endregion
                bodyA.ApplyImpulseAtWorldOffset(ref impulse, ref contact.R1);

                contactList[i] = contact;
            }
        }

        /// <summary>
        /// Fully exposed for convenience. Should be treated as readonly. Do not add or remove directly from this list.
        /// </summary>
        public ContactList ContactList {
            get { return contactList; }
        }

        internal int ContactCount {
            get { return contactList.Count; }
        }


        internal void Reset() {
            geomA = null;
            geomB = null;
            contactList.Clear();
            newContactList.Clear();
            mergedContactList.Clear();
        }

        internal void Collide() {
            newContactList.Clear();
            //_geometryA.Collide(_geometryB, _newContactList);
            Collide(GeomA, GeomB, newContactList);
            mergedContactList.Clear();

            for (int i = 0; i < newContactList.Count; i++) {
                //int index = _contactList.IndexOf(_newContactList[i]); **this is really commented out by author
                
                int index = contactList.IndexOfSafe(newContactList[i]);
                if (index > -1) {
                    //continuation of collision
                    Contact contact = newContactList[i];
                    contact.NormalImpulse = contactList[index].NormalImpulse;
                    contact.TangentImpulse = contactList[index].TangentImpulse;
                    mergedContactList.Add(contact);
                }
                else {
                    //first time collision
                    mergedContactList.Add(newContactList[i]);
                }
            }
            contactList.Clear();
            for (int i = 0; i < mergedContactList.Count; i++) {
                contactList.Add(mergedContactList[i]);
            }
        }

        #region Collide variables
        Feature feature;
        Vector2 localVertex;
        Vector2 vertRef;
        Matrix matrixInverseTemp;
        Matrix matrixTemp;
        #endregion
        private void Collide(Geom geometry1, Geom geometry2, ContactList contactList) {
            int vertexIndex = -1;
            //matrixInverseTemp = this.MatrixInverse;
            matrixTemp = geometry1.Matrix;
            for (int i = 0; i < geometry2.WorldVertices.Count; i++) {
                if (contactList.Count == physicsSimulator.maxContactsToDetect)
                {
                    break;
                }
                if (geomA.Grid == null) { break; }//grid can be null for "one-way" collision (points)

                vertexIndex += 1;
                vertRef = geometry2.WorldVertices[i];
                geometry1.TransformToLocalCoordinates(ref vertRef, out localVertex);

                if (!geomA.Grid.Intersect(ref localVertex, out feature)) { continue; }
                geomA.Grid.Intersect(ref localVertex, out feature);
                if (feature.Distance < 0f) {
                    geometry1.TransformNormalToWorld(ref feature.Normal, out feature.Normal);
                    Contact contact = new Contact(geometry2.WorldVertices[i], feature.Normal, feature.Distance);
                    contactList.Add(contact);
                }
            }

            matrixInverseTemp = Matrix.Invert(geometry2.Matrix);
            matrixTemp = geometry2.Matrix;

            for (int i = 0; i < geometry1.WorldVertices.Count; i++) {
                if (contactList.Count == physicsSimulator.maxContactsToDetect)
                {
                    break;
                }
                if (geomB.Grid == null) { break; }//grid can be null for "one-way" collision (points)

                vertexIndex += 1;

                vertRef = geometry1.WorldVertices[i];
                geometry2.TransformToLocalCoordinates(ref vertRef, out localVertex);
                if (!geomB.Grid.Intersect(ref localVertex, out feature)) { continue; }

                if (feature.Distance < 0f) {
                    geometry2.TransformNormalToWorld(ref feature.Normal, out feature.Normal);
                    feature.Normal = -feature.Normal;
                    Contact contact = new Contact(geometry1.WorldVertices[i], feature.Normal, feature.Distance);
                    contactList.Add(contact);
                }
            }

            //sort contact list by seperation (amount of penetration)
            contactList.Sort(contactComparer);

            //resolve deepest contacts first
            int contactCount = contactList.Count;
            if (contactList.Count > physicsSimulator.maxContactsToResolve)
            {
                contactList.RemoveRange(physicsSimulator.maxContactsToResolve, contactCount - physicsSimulator.maxContactsToResolve);
            }
        }
        
        private static int  CompareSeperation(Contact c1, Contact c2) {
            if (c1.Separation < c2.Separation) {
                return -1;
            }
            else if (c1.Separation == c2.Separation) {
                return 0;
            }
            else {
                return 1;
            }
        }

        private void InitializeContactLists(int maxContactsToDetect) {
            if (contactList == null) { contactList = new ContactList(maxContactsToDetect); }
            if (newContactList == null) { newContactList = new ContactList(maxContactsToDetect); }
            if (mergedContactList == null) { mergedContactList = new ContactList(maxContactsToDetect); }
        }

        public bool Equals(Arbiter2 other) {
            return (geomA == other.geomA) && (geomB == other.geomB);
        }

        public override bool Equals(object obj) {
            if (!(obj is Arbiter2)) { throw new ArgumentException("The object being compared must be of type 'Arbiter'"); }
            return Equals((Arbiter2)obj);
        }

        public override int GetHashCode() {
            return base.GetHashCode();
        }

        public static bool operator ==(Arbiter2 arbiter1, Arbiter2 arbiter2) {
            return arbiter1.Equals(arbiter2);
        }

        public static bool operator !=(Arbiter2 arbiter1, Arbiter2 arbiter2) {
            return !arbiter1.Equals(arbiter2);
        }

    }
}












